import os
from typing import Tuple

base_dir: str = os.path.dirname(os.path.abspath(__file__))
project_dir: str = os.path.join(base_dir, "..")

batch_size_default: int = 16

anomaly_crop_rows_default: int = 1
anomaly_crop_cols_default: int = 1
anomaly_imagenet_mean: Tuple[float, float, float] = (0.485, 0.456, 0.406)
anomaly_imagenet_std: Tuple[float, float, float] = (0.229, 0.224, 0.225)



import os
from typing import AnyStr, List


def get_filenames(image_folder) -> List[AnyStr]:
    """
    Collect all image paths from the folder and its subfolders

    Args:
        image_folder ([AnyStr]): Folder containing all images
    Return:
        [AnyStr]: List of all image paths
    """
    image_paths: List[AnyStr] = [
        os.path.join(root, img)
        for root, _, files in os.walk(image_folder)
        for img in files
        if img.endswith((".jpg", ".png"))
    ]
    return image_paths



class GridCrop:
    def __init__(self, rows=2, cols=2):
        """
        Args:
            rows (int): Number of rows to split the image into.
            cols (int): Number of columns to split the image into.
        """
        self.rows = rows
        self.cols = cols

    def __call__(self, img):
        width, height = img.size
        crop_width, crop_height = width // self.cols, height // self.rows

        crops = []

        # Split the image into a grid based on rows and columns
        for row in range(self.rows):
            for col in range(self.cols):
                left = col * crop_width
                upper = row * crop_height
                right = (col + 1) * crop_width
                lower = (row + 1) * crop_height

                crops.append(img.crop((left, upper, right, lower)))

        return crops



import logging
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple

import cv2
import numpy as np
import onnxruntime as ort
from PIL.Image import Image
from tqdm.auto import tqdm, trange

from anomalytda_client.constants import batch_size_default
from anomalytda_client.internal._model import BaseModel

logger: logging.Logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class DetectionResult:
    boxes: List[Tuple[int, int, int, int]]
    scores: List[float]
    class_indices: List[int]
    _model: "DetectionModel"

    def __iter__(self):
        return zip(self.boxes, self.scores, self.class_indices)

    @property
    def names(self) -> Dict[int, str]:
        return self._model.names


def _preprocess(batch_images: List[Image], input_size: Optional[int] = None) -> Tuple[List[np.ndarray], List[float]]:
    """
    Preprocess images for the model:
    - Convert image from PIL format to NumPy format.
    - Convert image from RGBA to RGB if needed.
    - Resize image to 640x640 pixels.
    - Normalize pixel values.
    - Change channel order to "channel-first".

    :param batch_images: List of images in PIL format.
    :param input_size: Image size in pixels.
    :return: List of preprocessed images in NumPy format.
    """
    input_size = input_size or 640
    preprocessed_images: List[np.ndarray] = []
    scales: List[float] = []
    for image in tqdm(batch_images, desc="Preprocessing images batch"):
        image_np = np.array(image if image.mode == "RGB" else image.convert("RGB"))

        shape = image_np.shape
        ratio = float(shape[0]) / shape[1]
        if ratio > 1:
            h = input_size
            w = int(h / ratio)
        else:
            w = input_size
            h = int(w * ratio)
        scale = float(h) / shape[0]
        scales.append(scale)

        image_np = cv2.resize(image_np, (w, h))

        det_image: np.ndarray = np.zeros((input_size, input_size, 3), dtype=np.uint8)
        det_image[:h, :w, :] = image_np

        det_image = det_image.astype(np.float32) / 255.0
        det_image = np.transpose(det_image, (2, 0, 1))
        preprocessed_images.append(det_image)
    preprocessed_images = np.stack(preprocessed_images, axis=0)
    return preprocessed_images, scales


def _postprocessing(
    outputs: np.ndarray, scale: float = 1.0, confidence_threshold: float = 0.1
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Post-process model outputs:
    - Extract class scores and bounding box coordinates.
    - Filter detections based on a confidence threshold.
    - Scale bounding box coordinates.

    :param outputs: Model outputs.
    :param scale: Scale for bounding box coordinates.
    :param confidence_threshold: Confidence threshold for filtering detections.
    :return: Scaled bounding box coordinates, filtered scores, and class indices.
    """
    class_scores = outputs[:, 4:]
    bounding_boxes = outputs[:, :4]

    max_scores = np.max(class_scores, axis=1)
    class_ids = np.argmax(class_scores, axis=1)

    valid_indices = np.where(max_scores >= confidence_threshold)[0]

    valid_boxes = bounding_boxes[valid_indices]
    valid_scores = max_scores[valid_indices]
    valid_class_ids = class_ids[valid_indices]

    left = np.maximum(0, ((valid_boxes[:, 0] - valid_boxes[:, 2] / 2) / scale).astype(int))
    top = np.maximum(0, ((valid_boxes[:, 1] - valid_boxes[:, 3] / 2) / scale).astype(int))
    width = (valid_boxes[:, 2] / scale).astype(int)
    height = (valid_boxes[:, 3] / scale).astype(int)

    scaled_boxes = np.stack([left, top, width, height], axis=1)

    return scaled_boxes, valid_scores, valid_class_ids


def _non_max_suppression_per_class(
    boxes: np.ndarray, scores: np.ndarray, class_indices: np.ndarray, conf_thres: float = 0.25, iou_thres: float = 0.45
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Perform Non-Maximum Suppression (NMS) per class:
    - Filter out bounding boxes with low confidence.
    - Apply NMS for each unique class.

    :param boxes: Bounding boxes.
    :param scores: Confidence scores.
    :param class_indices: Class indices.
    :param conf_thres: Confidence threshold for filtering.
    :param iou_thres: IoU threshold for NMS.
    :return: DetectionResult object with filtered boxes, scores, and class indices.
    """
    keep = scores >= conf_thres
    boxes, scores, class_indices = boxes[keep], scores[keep], class_indices[keep]

    if len(boxes) == 0:
        return np.array([]), np.array([]), np.array([])

    unique_classes = np.unique(class_indices)
    final_boxes = []
    final_scores = []
    final_class_indices = []

    for cls_ in unique_classes:
        cls_mask = class_indices == cls_
        cls_boxes = boxes[cls_mask]
        cls_scores = scores[cls_mask]
        cls_class_indices = class_indices[cls_mask]

        indices = np.argsort(cls_scores)[::-1]
        cls_boxes, cls_scores, cls_class_indices = cls_boxes[indices], cls_scores[indices], cls_class_indices[indices]

        while len(cls_boxes) > 0:
            current_box = cls_boxes[0]
            final_boxes.append(current_box)
            final_scores.append(cls_scores[0])
            final_class_indices.append(cls_class_indices[0])

            x1 = np.maximum(current_box[0], cls_boxes[1:, 0])
            y1 = np.maximum(current_box[1], cls_boxes[1:, 1])
            x2 = np.minimum(current_box[0] + current_box[2], cls_boxes[1:, 0] + cls_boxes[1:, 2])
            y2 = np.minimum(current_box[1] + current_box[3], cls_boxes[1:, 1] + cls_boxes[1:, 3])

            inter_area = np.maximum(0, x2 - x1) * np.maximum(0, y2 - y1)
            box_area = current_box[2] * current_box[3]
            other_areas = cls_boxes[1:, 2] * cls_boxes[1:, 3]

            iou = inter_area / (box_area + other_areas - inter_area)

            mask = iou <= iou_thres
            cls_boxes = cls_boxes[1:][mask]
            cls_scores = cls_scores[1:][mask]
            cls_class_indices = cls_class_indices[1:][mask]

    return np.array(final_boxes), np.array(final_scores), np.array(final_class_indices)


class DetectionModel(BaseModel):
    def __init__(
        self,
        anomaly_model_filepath: str,
        image_size: Optional[Tuple[int, int]] = None,
        batch_size: Optional[int] = None,
    ) -> None:
        """
        Constructor for the anomaly detection model.

        :param anomaly_model_filepath: Path to the anomaly model file.
        :param image_size: Model image size.
        :param batch_size: Batch size for processing images.
        """
        self.anomaly_model_filepath: str = anomaly_model_filepath
        self.image_size: Tuple[int, int] = image_size or (640, 640)
        self.batch_size: int = max(2, batch_size or batch_size_default)

    @property
    def session(self) -> ort.InferenceSession:
        """
        Creates and returns a session for performing inference using the ONNX model.

        :return: ONNX inference session.
        """
        if not hasattr(self, "_session"):
            self._session: ort.InferenceSession = self._load_model_session(model_name="best.onnx")

            # TODO: Delete debug information after test detection
            logger.debug("Model inputs:")
            for input in self._session.get_inputs():
                logger.debug(f"Name: {input.name}, Shape: {input.shape}, Type: {input.type}")

            logger.debug("Model outputs:")
            for input in self._session.get_outputs():
                logger.debug(f"Name: {input.name}, Shape: {input.shape}, Type: {input.type}")

        return self._session

    @property
    def names(self) -> Dict[int, str]:
        if not hasattr(self, "_names"):
            self._names: Dict[int, str] = eval(
                self.session.get_modelmeta().custom_metadata_map.get("names", "{0: ''}")
            )
        return self._names

    def inference(
        self, images_set: List[Image], conf: float = 0.1, iou: float = 0.45, *args, **kwargs
    ) -> List[DetectionResult]:
        """
        Performs inference on a set of images.

        :param images_set: List of images in PIL format.
        :param conf: Confidence threshold for filtering detections.
        :param iou: IoU threshold for non-maximum suppression.
        :return: List of inference results in NumPy format.
        """
        results: List[DetectionResult] = []
        for offset in trange(0, len(images_set), self.batch_size, desc="Predicting images batch"):
            input_images, scales = _preprocess(
                batch_images=images_set[offset : offset + self.batch_size], input_size=self.image_size[0]
            )
            outputs = self.session.run(None, {"images": input_images})[0]
            for i in trange(outputs.shape[0], desc="Postprocessing images batch"):
                boxes, scores, class_indices = _postprocessing(
                    outputs=np.transpose(np.squeeze(outputs[i])), scale=scales[i], confidence_threshold=conf
                )
                final_boxes, final_scores, final_class_indices = _non_max_suppression_per_class(
                    boxes=boxes, scores=scores, class_indices=class_indices, conf_thres=conf, iou_thres=iou
                )
                results.append(
                    DetectionResult(
                        boxes=final_boxes.tolist(),
                        scores=final_scores.tolist(),
                        class_indices=list(map(int, final_class_indices)),
                        _model=self,
                    )
                )
        return results



from anomalytda_client.internal._anomaly import AnomalyModel, get_anomalies
from anomalytda_client.internal._detection import DetectionModel, DetectionResult

__all__ = ["AnomalyModel", "DetectionModel", "DetectionResult", "get_anomalies"]



from abc import ABC, abstractmethod
from typing import Any, List
from zipfile import ZipFile

import onnxruntime as ort
from PIL.Image import Image


class BaseModel(ABC):
    def __init__(self, anomaly_model_filepath: str) -> None:
        self.anomaly_model_filepath: str = anomaly_model_filepath

    def _load_model_session(self, model_name: str) -> ort.InferenceSession:
        with ZipFile(self.anomaly_model_filepath) as zip_fn:
            with zip_fn.open(model_name) as fn:
                return ort.InferenceSession(fn.read())

    @abstractmethod
    def inference(self, images_set: List[Image], *args, **kwargs) -> List[Any]:
        pass



from typing import List, Tuple

import cv2
import numpy as np
from PIL import Image
from scipy.ndimage import find_objects, gaussian_filter, label, zoom
from tqdm.auto import tqdm


def min_max_norm(image: np.ndarray) -> np.ndarray:
    a_min, a_max = image.min(), image.max()
    return (image - a_min) / (a_max - a_min)


def cvt2heatmap(gray: np.ndarray) -> np.ndarray:
    heatmap = cv2.applyColorMap(np.uint8(gray), cv2.COLORMAP_JET)  # type: ignore
    return heatmap


def show_cam_on_image(img: np.uint8, anomaly_map: np.ndarray) -> np.uint8:
    cam = np.float32(anomaly_map) / 255 + np.float32(img) / 255
    cam = cam / np.max(cam)
    return np.uint8(255 * cam)


def zoom_anomaly_maps(combined_anomaly_maps: List[np.ndarray], new_size: Tuple[int, int]) -> List[np.ndarray]:
    a_map_list = []

    for combined_map in combined_anomaly_maps:
        expanded_combined_map = np.expand_dims(np.expand_dims(combined_map, axis=0), axis=0)
        zoom_factors = (
            new_size[0] / expanded_combined_map.shape[2],  # height zoom
            new_size[1] / expanded_combined_map.shape[3],  # width zoom
        )  
        combined_map = zoom(
            expanded_combined_map, (1, 1, zoom_factors[1], zoom_factors[0]), order=1
        )  # order=1 for bilinear

        # Store results
        a_m = combined_map[0, 0, :, :]
        a_map_list.append(a_m)
    return a_map_list


def get_images_amaps_score(
    images_set: List[Image.Image], amap_list: List[np.ndarray]
) -> Tuple[List[Tuple[np.ndarray, np.ndarray]], List[float]]:
    img_list: List[Tuple[np.ndarray, np.ndarray]] = []
    score_list: List[float] = []

    for i, amap in tqdm(
        enumerate(zoom_anomaly_maps(amap_list, images_set[0].size)),
        desc="Render anomalies on images",
        total=len(amap_list),
        unit="image",
    ):
        score_list.append(amap.max())
        anomaly_map = gaussian_filter(amap, sigma=4)
        ano_map = min_max_norm(anomaly_map)
        ano_map = (ano_map * -1) + 1
        ano_map = cvt2heatmap(ano_map * 255)
        img_vis = np.array(images_set[i].resize(ano_map.shape[:-1][::-1], Image.Resampling.LANCZOS))
        ano_map = show_cam_on_image(img_vis, ano_map)  # type: ignore
        img_list.append((img_vis, ano_map))

    return img_list, score_list


def find_bounding_boxes(amap: np.ndarray, image_size: Tuple[int, int], threshold: float = 0.5):
    sx, sy = np.shape(amap)
    zoom_factors: Tuple[float, float] = (image_size[0] / sx, image_size[1] / sy)
    amap_zoomed: np.ndarray = zoom(amap, zoom_factors, order=1)  # order=1 for bilinear

    binary_map = amap_zoomed > threshold

    # Label connected components
    labeled_array, _ = label(binary_map)

    # Find bounding boxes around hot spots
    hotspot_slices = find_objects(labeled_array)

    boxes: List[Tuple[int, int, int, int]] = [
        (
            hotspot_slice[1].start,  # x_min
            hotspot_slice[0].start,  # y_min
            hotspot_slice[1].stop,  # x_max,
            hotspot_slice[0].stop,  # y_max
        )
        for hotspot_slice in hotspot_slices
        if hotspot_slice is not None
    ]

    return boxes



from typing import Generator, List, Tuple

import numpy as np
from PIL import Image

from anomalytda_client.constants import anomaly_imagenet_mean, anomaly_imagenet_std
from anomalytda_client.files import get_filenames
from anomalytda_client.internal._grid_crop import GridCrop


def get_test_transforms_cpu(
    image_folder: str,
    image_size: Tuple[int, int] = (224, 224),
    crop_rows: int = 2,
    crop_cols: int = 2,
    batch_size: int = 32,
) -> Generator[np.ndarray, None, None]:
    """
    Generator to process images in batches, applying grid cropping, resizing, and normalization for ONNX inference.

    Args:
        image_folder (str): Path to the folder containing images.
        image_size (tuple): Target image size for resizing.
        crop_rows (int): Number of rows for grid cropping.
        crop_cols (int): Number of columns for grid cropping.
        batch_size (int): Number of images in each batch.

    Yields:
        numpy.ndarray: Batch of processed images, shape (batch_size, num_crops, 224, 224, 3).
    """

    batch = []
    for idx, image_path in enumerate(get_filenames(image_folder=image_folder)):
        # Load image
        image = Image.open(image_path)

        # Stack crops into a single array for this image and add to batch
        batch.append(
            _transform_image(image=image, image_size=image_size, crop_rows=crop_rows, crop_cols=crop_cols)
        )  # shape: (num_crops, 224, 224, 3)

        # Yield batch when it reaches the batch size
        if (idx + 1) % batch_size == 0:
            yield np.transpose(np.stack(batch, axis=0), (0, 1, 4, 2, 3))  # shape: (batch_size, num_crops, 3, 224, 224)
            batch = []

    # Yield any remaining images in the last batch
    if batch:
        yield np.transpose(np.stack(batch, axis=0), (0, 1, 4, 2, 3))


def get_test_transforms_cpu_from_images(
    images_set: List[Image.Image],
    image_size: Tuple[int, int] = (224, 224),
    crop_rows: int = 2,
    crop_cols: int = 2,
    batch_size: int = 32,
) -> Generator[np.ndarray, None, None]:
    for offset in range(0, len(images_set), batch_size):
        yield np.transpose(
            np.stack(
                [
                    _transform_image(image=image, image_size=image_size, crop_rows=crop_rows, crop_cols=crop_cols)
                    for image in images_set[offset : offset + batch_size]
                ],
                axis=0,
            ),
            (0, 1, 4, 2, 3),
        )  # shape: (batch_size, num_crops, 3, 224, 224)


def _transform_image(
    image: Image.Image, image_size: Tuple[int, int] = (224, 224), crop_rows: int = 2, crop_cols: int = 2
) -> np.ndarray:
    grid_crop = GridCrop(rows=crop_rows, cols=crop_cols)

    if image.mode != "RGB":
        image = image.convert(mode="RGB")

    # Apply GridCrop to the image
    crops = grid_crop(image)

    # Resize and normalize each crop
    processed_crops = []
    for crop in crops:
        resized_crop = crop.resize(image_size, Image.Resampling.LANCZOS)
        np_crop = np.array(resized_crop).astype(np.float32) / 255.0
        normalized_crop = (np_crop - anomaly_imagenet_mean) / anomaly_imagenet_std
        processed_crops.append(normalized_crop)

    # Stack crops into a single array for this image and add to batch
    return np.stack(processed_crops, axis=0)  # shape: (num_crops, 224, 224, 3)



import math
import os
from typing import List, Optional, Tuple

import numpy as np
import onnxruntime as ort
from PIL.Image import Image
from scipy.ndimage import gaussian_filter
from tqdm.auto import tqdm, trange

from anomalytda_client.constants import anomaly_crop_cols_default, anomaly_crop_rows_default, batch_size_default
from anomalytda_client.files import get_filenames
from anomalytda_client.internal._model import BaseModel
from anomalytda_client.internal._transforms import get_test_transforms_cpu, get_test_transforms_cpu_from_images
from anomalytda_client.tools import cvt2heatmap, min_max_norm, show_cam_on_image, zoom_anomaly_maps


def get_model_session(models_root: str, model_name: str) -> ort.InferenceSession:
    model_filepath: str = os.path.join(models_root, model_name)
    assert os.path.isfile(model_filepath)
    return ort.InferenceSession(model_filepath)


def combine_images_cpu(images: List[np.ndarray], rows: int, cols: int) -> np.ndarray:
    if images[0].ndim == 2:  # single channel images
        combined_rows = []
        for r in range(rows):
            # Concatenate images horizontally to form a single row
            row_images = images[r * cols : (r + 1) * cols]
            combined_row = np.concatenate(row_images, axis=-1)  # numpy version
            combined_rows.append(combined_row)

        # Now, concatenate all rows vertically
        combined_image = np.concatenate(combined_rows, axis=-2)  # numpy version
    else:  # multi channel images
        combined_rows = []
        for r in range(rows):
            # Concatenate the images in this row along the width (axis=1, which is dim=-2 for tensors)
            row_images = images[r * cols : (r + 1) * cols]
            combined_row = np.concatenate(row_images, axis=1)  # numpy version
            combined_rows.append(combined_row)

        # Now, concatenate the combined rows along the height (axis=0, which is dim=-3 for tensors)
        combined_image = np.concatenate(combined_rows, axis=0)  # numpy version

    return combined_image


def get_combined_anomaly_maps(
    fs_list: np.ndarray, ft_list: np.ndarray, crop_rows: int, crop_cols: int
) -> Tuple[List[np.ndarray], List[np.ndarray]]:
    combined_anomaly_maps: List[np.ndarray] = []
    sim_vector_list: List[np.ndarray] = []
    fs = fs_list[:, :, :, :]
    ft = ft_list[:, :, :, :]

    dot_product = np.sum(fs * ft, axis=1)  # shape: (72, 14, 14)
    fs_norm = np.linalg.norm(fs, axis=1)  # shape: (72, 14, 14)
    ft_norm = np.linalg.norm(ft, axis=1)  # shape: (72, 14, 14)
    cosine_similarity = dot_product / (fs_norm * ft_norm + 1e-8)  # Add small value to avoid division by zero
    a_map = 1 - cosine_similarity

    sim_vector = a_map.copy()

    a_map = np.expand_dims(a_map, axis=1)  # numpy version

    # Combine the anomaly maps
    for i in range(len(ft_list) // (crop_rows * crop_cols)):
        images = [a_map[i * crop_rows * crop_cols + j][0] for j in range(crop_rows * crop_cols)]
        combined_anomaly_maps.append(combine_images_cpu(images, crop_rows, crop_cols))
        sim_vector_list.append(
            sim_vector[i * (crop_rows * crop_cols) : (i + 1) * (crop_rows * crop_cols), :, :]
        )  # Store sim_vector for the corresponding set of 4

    return combined_anomaly_maps, sim_vector_list


def cal_anomaly_map_cpu(
    fs_list, ft_list, image_size, crop_rows, crop_cols, out_size=224, amap_mode="mul"
) -> Tuple[List[np.ndarray], List[np.ndarray]]:
    # Combine the anomaly maps
    combined_anomaly_maps, sim_vector_list = get_combined_anomaly_maps(
        fs_list=fs_list, ft_list=ft_list, crop_rows=crop_rows, crop_cols=crop_cols
    )

    a_map_list: List[np.ndarray] = zoom_anomaly_maps(
        combined_anomaly_maps=combined_anomaly_maps, new_size=(image_size[0] * crop_rows, image_size[1] * crop_cols)
    )

    return a_map_list, sim_vector_list


def get_anomalies(
    models_root: str,
    test_root: str,
    crop_rows: Optional[int] = None,
    crop_cols: Optional[int] = None,
    batch_size: Optional[int] = None,
) -> Tuple[List[Tuple[np.uint8, np.ndarray]], List[float], List[str]]:
    crop_rows = crop_rows or anomaly_crop_rows_default
    crop_cols = crop_cols or anomaly_crop_cols_default
    batch_size = batch_size or batch_size_default
    img_list: List[Tuple[np.uint8, np.ndarray]] = []
    score_list: List[float] = []

    encoder_session: ort.InferenceSession = get_model_session(models_root=models_root, model_name="encoder.onnx")
    bn_session: ort.InferenceSession = get_model_session(models_root=models_root, model_name="bn.onnx")
    decoder_session: ort.InferenceSession = get_model_session(models_root=models_root, model_name="decoder.onnx")

    image_size = tuple(encoder_session._inputs_meta[0].shape[-2:])

    filenames: List[str] = get_filenames(image_folder=test_root)

    for img in tqdm(
        get_test_transforms_cpu(test_root, image_size, crop_rows, crop_cols, batch_size),
        desc="Detecting anomalies",
        total=math.ceil(len(filenames) / batch_size),
        unit="batch",
    ):
        # Reshape the input: [batch_size, 4, 3, 224, 224] -> [batch_size * 4, 3, 224, 224]
        batch_size, num_quadrants, channels, height, width = img.shape
        img = img.reshape(batch_size * num_quadrants, channels, height, width)
        numpy_img = img.astype(np.float32)
        inputs = encoder_session.run(None, {"input": numpy_img})
        bn_outputs = bn_session.run(None, {"input": inputs[0], "input.13": inputs[1], "onnx::Concat_2": inputs[2]})
        outputs = decoder_session.run(None, {"input": bn_outputs[0]})

        amap_list, _ = cal_anomaly_map_cpu(
            inputs[-1],
            outputs[-1],
            out_size=img.shape[-1],
            image_size=image_size,
            crop_rows=crop_rows,
            crop_cols=crop_cols,
            amap_mode="acc",
        )

        # Prepare the images for visualization by permuting the axes
        imgs_vis = img.transpose(0, 2, 3, 1)  # numpy

        imgs_vis_combined = []
        for i in trange(len(imgs_vis) // num_quadrants, desc="Combine images", unit="image"):
            # Extract the corresponding quadrant images for this batch
            images = [imgs_vis[i * num_quadrants + j] for j in range(num_quadrants)]

            # Combine the images into a larger grid (e.g., 2x2, 3x3, etc.)
            combined_map = combine_images_cpu(images, crop_rows, crop_cols)

            # Append the combined image to the list for further processing or visualization
            imgs_vis_combined.append(combined_map)

        for i, amap in tqdm(
            enumerate(amap_list), desc="Render anomalies on images", total=len(amap_list), unit="image"
        ):
            score_list.append(amap.max())
            anomaly_map = gaussian_filter(amap, sigma=4)
            ano_map = min_max_norm(anomaly_map)
            ano_map = (ano_map * -1) + 1
            ano_map = cvt2heatmap(ano_map * 255)
            img_vis_c = imgs_vis_combined[i]  # cpu version
            img_vis = np.uint8(min_max_norm(img_vis_c) * 255)  # numpu
            # img_vis = cv2.cvtColor(img_vis_c, cv2.COLOR_BGR2RGB)
            ano_map = show_cam_on_image(img_vis, ano_map)  # type: ignore
            img_list.append((img_vis, ano_map))

    return img_list, score_list, filenames


class AnomalyModel(BaseModel):
    def __init__(
        self,
        anomaly_model_filepath: str,
        crop_rows: Optional[int] = None,
        crop_cols: Optional[int] = None,
        batch_size: Optional[int] = None,
    ) -> None:
        """
        Initial model for anomaly inference from pretrained model on DataRefiner

        :param anomaly_model_filepath: Path to pretrained anomaly model
        :param crop_rows: Number of rows for crop images
        :param crop_cols: Number of cols for crop images
        :param batch_size: Batch size for inference process
        """
        self.anomaly_model_filepath: str = anomaly_model_filepath
        self.crop_rows: int = crop_rows or anomaly_crop_rows_default
        self.crop_cols: int = crop_cols or anomaly_crop_cols_default
        self.batch_size: int = batch_size or batch_size_default

        assert os.path.isfile(self.anomaly_model_filepath), "Not found anomaly model"
        assert self.crop_rows == self.crop_cols, "Must be equal crop row and cols"

    @property
    def encoder_session(self) -> ort.InferenceSession:
        if not hasattr(self, "_encoder_session"):
            self._encoder_session: ort.InferenceSession = self._load_model_session("encoder.onnx")
        return self._encoder_session

    @property
    def bn_session(self) -> ort.InferenceSession:
        if not hasattr(self, "_bn_session"):
            self._bn_session: ort.InferenceSession = self._load_model_session("bn.onnx")
        return self._bn_session

    @property
    def decoder_session(self) -> ort.InferenceSession:
        if not hasattr(self, "_decoder_session"):
            self._decoder_session: ort.InferenceSession = self._load_model_session("decoder.onnx")
        return self._decoder_session

    def inference(self, images_set: List[Image]) -> List[np.ndarray]:
        image_size: Tuple[int, int] = tuple(self.encoder_session._inputs_meta[0].shape[-2:])

        combined_anomaly_maps: List[np.ndarray] = []
        for img in tqdm(
            get_test_transforms_cpu_from_images(
                images_set, image_size, self.crop_rows, self.crop_cols, self.batch_size
            ),
            desc="Detecting anomalies",
            total=math.ceil(len(images_set) / self.batch_size),
            unit="batch",
        ):
            # Reshape the input: [batch_size, 4, 3, 224, 224] -> [batch_size * 4, 3, 224, 224]
            batch_size, num_quadrants, channels, height, width = img.shape
            img = img.reshape(batch_size * num_quadrants, channels, height, width)
            numpy_img = img.astype(np.float32)
            inputs = self.encoder_session.run(None, {"input": numpy_img})
            bn_outputs = self.bn_session.run(
                None, {"input": inputs[0], "input.13": inputs[1], "onnx::Concat_2": inputs[2]}
            )
            outputs = self.decoder_session.run(None, {"input": bn_outputs[0]})

            _combined_anomaly_maps, _ = get_combined_anomaly_maps(
                inputs[-1], outputs[-1], crop_rows=self.crop_rows, crop_cols=self.crop_cols
            )
            combined_anomaly_maps.extend(_combined_anomaly_maps)
        return combined_anomaly_maps
